/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: marray.h 23 2003-10-28 02:24:16Z drory $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2002 Niv Drory <drory@usm.uni-muenchen.de>
 *                         Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */


// ====================================================================

#ifndef __LTL_MARRAY__
#define __LTL_MARRAY__

#include <cmath>

#include <iostream>
#include <list>

// ====================================================================

// little trick to avoid files being included separetely
#define __LTL_IN_FILE_MARRAY__
// this is #undef'd at the end of this file ...

// ====================================================================


// mandatory includes with own namespace ltl declaration
// (so that they can be used independently of marray.h)
#include <ltl/config.h>

#include <ltl/misc/mdebug.h>

//#include <ltl/fvector.h>
#include <ltl/misc/exceptions.h>
#include <ltl/misc/memblock.h>
#include <ltl/misc/range.h>

// ====================================================================

#ifdef LTL_USING_NAMESPACE
using std::list;

LTL_BEGIN_NAMESPACE
#endif

// these must be included from within the namespace ltl
#include <ltl/marray/shape.h>
#include <ltl/marray/slice.h>
#include <ltl/misc/staticinit.h>
#include <ltl/marray/simplevec.h>

// ====================================================================

#include <ltl/marray/marray_iter.h>
#include <ltl/marray/expr.h>
#include <ltl/marray/expr_ops.h>
#include <ltl/marray/eval.h>
#include <ltl/marray/index_iter.h>

#include <ltl/marray/merge.h>
#include <ltl/marray/where.h>

// ====================================================================


// N-dimensional array storing objects of type T.
//
template<class T, int N>
class MArray
{

   public:
      friend class MArrayIter<T,N>;
      friend class MArrayIterConst<T,N>;
      friend class ListInitializationSwitch<MArray<T,N> >;

      typedef MArrayIter<T,N>       Iterator;
      typedef MArrayIterConst<T,N>  ConstIterator;

      // index iterator for where() (indexing arbitrary subsets)
      typedef IndexIter<T,N>        IndexIterator;
      typedef FixedVector<int,N>    IndexV;
      typedef IndexList<N>          IndexSet;

      // STL definitions
      typedef T                     value_type;
      typedef T&                    reference;
      typedef const T&              const_reference;
      typedef T*                    pointer;
      typedef const T*              const_pointer;
      typedef size_t                size_type;
      typedef MArrayIter<T,N>       iterator;
      typedef MArrayIterConst<T,N>  const_iterator;



      // CONSTRUCTORS
      MArray()
      {
         memBlock_=NULL;
      }


      // Range arguments
      MArray( const Range& r1 )
            : shape_( r1 )
      {
         setupMemory();
      }

      MArray( const Range& r1, const Range& r2 )
            : shape_( r1, r2 )
      {
         setupMemory();
      }

      MArray( const Range& r1, const Range& r2, const Range& r3 )
            : shape_(r1,r2,r3)
      {
         setupMemory();
      }

      MArray( const Range& r1, const Range& r2, const Range& r3,
              const Range& r4 )
            : shape_(r1,r2,r3,r4)
      {
         setupMemory();
      }

      MArray( const Range& r1, const Range& r2, const Range& r3,
              const Range& r4, const Range& r5 )
            : shape_(r1,r2,r3,r4,r5)
      {
         setupMemory();
      }


      // int arguments, base defaults to 1
      MArray( const int r1 )
            : shape_( r1 )
      {
         setupMemory();
      }

      MArray( const int r1, const int r2 )
            : shape_( r1, r2 )
      {
         setupMemory();
      }

      MArray( const int r1, const int r2, const int r3 )
            : shape_( r1, r2, r3 )
      {
         setupMemory();
      }

      MArray( const int r1, const int r2, const int r3, const int r4 )
            : shape_( r1, r2, r3, r4 )
      {
         setupMemory();
      }

      MArray( const int r1, const int r2, const int r3, const int r4,
              const int r5)
            : shape_( r1, r2, r3, r4, r5 )
      {
         setupMemory();
      }


      // copy constructor. only makes a reference !!!
      MArray( const MArray<T, N>& other )
            : memBlock_(0)
      {
         makeReference( other );
      }

      // constructors for pure subarrays (rank preserved)
      // other array's data is referenced, NOT copied
      // ( use operator=() for copy)!

      // missing arguments are treated as Range::all() !
      MArray( const MArray<T, N>& other, const Range& r0 )
            : memBlock_(0)
      {
         subarray( other, r0 );
      }

      MArray( const MArray<T, N>& other,
              const Range& r1, const Range& r2 )
            : memBlock_(0)
      {
         subarray( other, r1, r2 );
      }

      MArray( const MArray<T, N>& other,
              const Range& r1, const Range& r2, const Range& r3 )
            : memBlock_(0)
      {
         subarray( other, r1, r2, r3 );
      }

      MArray( const MArray<T, N>& other,
              const Range& r1, const Range& r2, const Range& r3,
              const Range& r4 )
            : memBlock_(0)
      {
         subarray( other, r1, r2, r3, r4 );
      }

      MArray( const MArray<T, N>& other,
              const Range& r1, const Range& r2, const Range& r3,
              const Range& r4, const Range& r5 )
            : memBlock_(0)
      {
         subarray( other, r1, r2, r3, r4, r5 );
      }


      // constructor for mixed slicing (int, rank reducing) and Range arguments
      // other array's data is referenced, NOT copied
      // ( use operator=() for copy)!
      //
      // not for public use, user operator()
      template<int N2, class R1, class R2, class R3, class R4, class R5>
      MArray( const MArray<T,N2>& other,
              R1 r1, R2 r2, R3 r3, R4 r4, R5 r5 )
            : memBlock_(0)
      {
         slicearray( other, r1, r2, r3, r4, r5 ) ;
      }

      // construct from preexisting data
      MArray( T *data, int *dims );

      // construct from shape
      // to get an equally shaped array do MArray<A,B> B(A.shape());
      MArray( const Shape<N>* s );

      // construct from expression
      template<class Expr>
      MArray( TExpr<Expr,N> e );


      // DESTRUCTOR
      virtual ~MArray()
      {
         if( memBlock_ )
            memBlock_->removeReference();
      }


      // FREE ASSOCIATED MEMORY
      // for freeing the memory used to store the data before the MArray goes
      // out of scope (before the destructor is being called). Can help keeping
      // code cleaner without using pointers to MArray.
      void free()
      {
         if( memBlock_ )
            memBlock_->removeReference();
         memBlock_ = NULL;
         data_     = NULL;
      }

      // OPERATOR =
      // other's data is copied, own data overwritten
      // Arrays have to have conformable shapes, no automatic resize takes
      // place. Use realloc() instead!
      //
      MArray<T,N>& operator=( const MArray<T,N>& other )
      {
         copy( other );
         return *this;
      }

      template<class T2>
      MArray<T,N>& operator=( const MArray<T2,N>& other )
      {
         copy( other );
         return *this;
      }

      // operator= for expressions
      // see above for realloc()
      template<class Expr>
      MArray<T,N>& operator=( TExpr<Expr,N> e );

      // assigns t to all elements
      // a bit more comlicated since we have to discriminate between
      // A = 3; and A = 1, 2, 3, 4;
      // which is done using ListInitializationSwitch which either calls
      // ListInitializer or MArray::fill().
      ListInitializationSwitch< MArray<T,N> > operator=( T x )
      {
         return ListInitializationSwitch< MArray<T,N> >( *this, x );
      }


      // OVERLOADED X= OERATORS
      template<class T2>
      MArray<T,N>& operator+=( const MArray<T2,N>& a );
      template<class T2>
      MArray<T,N>& operator-=( const MArray<T2,N>& a );
      template<class T2>
      MArray<T,N>& operator*=( const MArray<T2,N>& a );
      template<class T2>
      MArray<T,N>& operator/=( const MArray<T2,N>& a );
      template<class T2>
      MArray<T,N>& operator%=( const MArray<T2,N>& a );
      template<class T2>
      MArray<T,N>& operator&=( const MArray<T2,N>& a );
      template<class T2>
      MArray<T,N>& operator|=( const MArray<T2,N>& a );
      template<class T2>
      MArray<T,N>& operator^=( const MArray<T2,N>& a );
      template<class T2>
      MArray<T,N>& operator<<=( const MArray<T2,N>& a );
      template<class T2>
      MArray<T,N>& operator>>=( const MArray<T2,N>& a );

      // the same with expression template rhs
      template<class Expr>
      MArray<T,N>& operator+=( TExpr<Expr,N> a );
      template<class Expr>
      MArray<T,N>& operator-=( TExpr<Expr,N> a );
      template<class Expr>
      MArray<T,N>& operator*=( TExpr<Expr,N> a );
      template<class Expr>
      MArray<T,N>& operator/=( TExpr<Expr,N> a );
      template<class Expr>
      MArray<T,N>& operator%=( TExpr<Expr,N> a );
      template<class Expr>
      MArray<T,N>& operator&=( TExpr<Expr,N> a );
      template<class Expr>
      MArray<T,N>& operator|=( TExpr<Expr,N> a );
      template<class Expr>
      MArray<T,N>& operator^=( TExpr<Expr,N> a );
      template<class Expr>
      MArray<T,N>& operator<<=( TExpr<Expr,N> a );
      template<class Expr>
      MArray<T,N>& operator>>=( TExpr<Expr,N> a );

      // scalars
      MArray<T,N>& operator+=( const T a );
      MArray<T,N>& operator-=( const T a );
      MArray<T,N>& operator*=( const T a );
      MArray<T,N>& operator/=( const T a );
      MArray<T,N>& operator%=( const T a );
      MArray<T,N>& operator&=( const T a );
      MArray<T,N>& operator|=( const T a );
      MArray<T,N>& operator^=( const T a );
      MArray<T,N>& operator<<=( const T a );
      MArray<T,N>& operator>>=( const T a );


      // ARRAY ACCESS (INDEXING) via OPERATOR ()

      // integer arguments, return (reference to) element
      T  operator()( const int i1 ) const
      {
         ASSERT_DIM(1);
         CHECK_BOUNDS1(i1);
         return data_[i1*shape_.stride_[0]];
      }
      T& restrict_ operator()( const int i1 )
      {
         ASSERT_DIM(1);
         CHECK_BOUNDS1(i1);
         return data_[i1*shape_.stride_[0]];
      }

      T  operator()( const int i1, const int i2 ) const
      {
         ASSERT_DIM(2);
         CHECK_BOUNDS2(i1,i2);
         return data_[i1*shape_.stride_[0] + i2*shape_.stride_[1]];
      }
      T& restrict_ operator()( const int i1, const int i2 )
      {
         ASSERT_DIM(2);
         CHECK_BOUNDS2(i1,i2);
         return data_[i1*shape_.stride_[0] + i2*shape_.stride_[1]];
      }

      T  operator()( const int i1, const int i2,
                     const int i3 ) const
      {
         ASSERT_DIM(3);
         CHECK_BOUNDS3(i1,i2,i3);
         return data_[i1*shape_.stride_[0] + i2*shape_.stride_[1] +
                      i3*shape_.stride_[2]];
      }
      T& restrict_ operator()( const int i1, const int i2,
                               const int i3 )
      {
         ASSERT_DIM(3);
         CHECK_BOUNDS3(i1,i2,i3);
         return data_[i1*shape_.stride_[0] + i2*shape_.stride_[1] +
                      i3*shape_.stride_[2]];
      }

      T  operator()( const int i1, const int i2,
                     const int i3, const int i4 ) const
      {
         ASSERT_DIM(4);
         CHECK_BOUNDS4(i1,i2,i3,i4);
         return data_[i1*shape_.stride_[0] + i2*shape_.stride_[1] +
                      i3*shape_.stride_[2] + i4*shape_.stride_[3] ];
      }
      T& restrict_ operator()( const int i1, const int i2,
                               const int i3, const int i4 )
      {
         ASSERT_DIM(4);
         CHECK_BOUNDS4(i1,i2,i3,i4);
         return data_[i1*shape_.stride_[0] + i2*shape_.stride_[1] +
                      i3*shape_.stride_[2] + i4*shape_.stride_[3] ];
      }

      T  operator()( const int i1, const int i2,
                     const int i3, const int i4,
                     const int i5 ) const
      {
         ASSERT_DIM(5);
         CHECK_BOUNDS5(i1,i2,i3,i4,i5);
         return data_[i1*shape_.stride_[0] + i2*shape_.stride_[1] +
                      i3*shape_.stride_[2] + i4*shape_.stride_[3] +
                      i5*shape_.stride_[4] ];
      }
      T& restrict_ operator()( const int i1, const int i2,
                               const int i3, const int i4,
                               const int i5 )
      {
         ASSERT_DIM(5);
         CHECK_BOUNDS5(i1,i2,i3,i4,i5);
         return data_[i1*shape_.stride_[0] + i2*shape_.stride_[1] +
                      i3*shape_.stride_[2] + i4*shape_.stride_[3] +
                      i5*shape_.stride_[4] ];
      }



      // FixedVector argument, return (reference to) element
      T operator()( const FixedVector<int,1>& i ) const
      {
         ASSERT_DIM(1);
         CHECK_BOUNDS1(i(1));
         return data_[i(1)*shape_.stride_[0]];
      }

      T& restrict_ operator()( const FixedVector<int,1>& i )
      {
         ASSERT_DIM(1);
         CHECK_BOUNDS1(i(1));
         return data_[i(1)*shape_.stride_[0]];
      }

      T operator()( const FixedVector<int,2>& i ) const
      {
         ASSERT_DIM(2);
         CHECK_BOUNDS2(i(1),i(2));
         return data_[i(1)*shape_.stride_[0]+i(2)*shape_.stride_[1]];
      }

      T& restrict_ operator()( const FixedVector<int,2>& i )
      {
         ASSERT_DIM(2);
         CHECK_BOUNDS2(i(1),i(2));
         return data_[i(1)*shape_.stride_[0]+i(2)*shape_.stride_[1]];
      }

      T operator()( const FixedVector<int,3>& i ) const
      {
         ASSERT_DIM(3);
         CHECK_BOUNDS3(i(1),i(2),i(3));
         return data_[i(1)*shape_.stride_[0]+i(2)*shape_.stride_[1]+
                      i(3)*shape_.stride_[2]];
      }

      T& restrict_ operator()( const FixedVector<int,3>& i )
      {
         ASSERT_DIM(3);
         CHECK_BOUNDS3(i(1),i(2),i(3));
         return data_[i(1)*shape_.stride_[0]+i(2)*shape_.stride_[1]+
                      i(3)*shape_.stride_[2]];
      }

      T operator()( const FixedVector<int,4>& i ) const
      {
         ASSERT_DIM(4);
         CHECK_BOUNDS4(i(1),i(2),i(3),i(4));
         return data_[i(1)*shape_.stride_[0]+i(2)*shape_.stride_[1]+
                      i(3)*shape_.stride_[2]+i(4)*shape_.stride_[3]];
      }

      T& restrict_ operator()( const FixedVector<int,4>& i )
      {
         ASSERT_DIM(4);
         CHECK_BOUNDS4(i(1),i(2),i(3),i(4));
         return data_[i(1)*shape_.stride_[0]+i(2)*shape_.stride_[1]+
                      i(3)*shape_.stride_[2]+i(4)*shape_.stride_[3]];
      }

      T operator()( const FixedVector<int,5>& i ) const
      {
         ASSERT_DIM(5);
         CHECK_BOUNDS5(i(1),i(2),i(3),i(4),i(5));
         return data_[i(1)*shape_.stride_[0]+i(2)*shape_.stride_[1]+
                      i(3)*shape_.stride_[2]+i(4)*shape_.stride_[3]+
                      i(5)*shape_.stride_[4]];
      }

      T& restrict_ operator()( const FixedVector<int,5>& i )
      {
         ASSERT_DIM(5);
         CHECK_BOUNDS5(i(1),i(2),i(3),i(4),i(5));
         return data_[i(1)*shape_.stride_[0]+i(2)*shape_.stride_[1]+
                      i(3)*shape_.stride_[2]+i(4)*shape_.stride_[3]+
                      i(5)*shape_.stride_[4]];
      }



      // Range arguments: return subarray
      MArray<T,N> operator()( const Range& r1 ) const
      {
         ASSERT_DIM(1);
         CHECK_RANGE(r1,1);
         return MArray<T,N>( *this, r1 );
      }

      MArray<T,N> operator()( const Range& r1, const Range& r2 ) const
      {
         ASSERT_DIM(2);
         CHECK_RANGE(r1,1);
         CHECK_RANGE(r2,2);
         return MArray<T,N>( *this, r1, r2 );
      }

      MArray<T,N> operator()( const Range& r1, const Range& r2,
                              const Range& r3 ) const
      {
         ASSERT_DIM(3);
         CHECK_RANGE(r1,1);
         CHECK_RANGE(r2,2);
         CHECK_RANGE(r3,3);
         return MArray<T,N>( *this, r1, r2, r3 );
      }

      MArray<T,N> operator()( const Range& r1, const Range& r2,
                              const Range& r3, const Range& r4 ) const
      {
         ASSERT_DIM(4);
         CHECK_RANGE(r1,1);
         CHECK_RANGE(r2,2);
         CHECK_RANGE(r3,3);
         CHECK_RANGE(r4,4);
         return MArray<T,N>( *this, r1, r2, r3, r4 );
      }

      MArray<T,N> operator()( const Range& r1, const Range& r2,
                              const Range& r3, const Range& r4,
                              const Range& r5 ) const
      {
         ASSERT_DIM(5);
         CHECK_RANGE(r1,1);
         CHECK_RANGE(r2,2);
         CHECK_RANGE(r3,3);
         CHECK_RANGE(r4,4);
         CHECK_RANGE(r5,5);
         return MArray<T,N>( *this, r1, r2, r3, r4, r5 );
      }

      // mixed interger and Range arguments, return subarray/slice
      template<class T1, class T2>
      typename SliceCounter<T,T1,T2>::MArraySlice
      operator()(T1 r1, T2 r2) const
      {
         return typename SliceCounter<T,T1,T2>::
                MArraySlice(*this, r1, r2,
                            NoArgument(), NoArgument(), NoArgument());
      }

      template<class T1, class T2, class T3>
      typename SliceCounter<T,T1,T2,T3>::MArraySlice
      operator()(T1 r1, T2 r2, T3 r3) const
      {
         return typename SliceCounter<T,T1,T2,T3>::
                MArraySlice(*this, r1, r2, r3, NoArgument(), NoArgument() );
      }

      template<class T1, class T2, class T3, class T4>
      typename SliceCounter<T,T1,T2,T3,T4>::MArraySlice
      operator()(T1 r1, T2 r2, T3 r3, T4 r4) const
      {
         return typename SliceCounter<T,T1,T2,T3,T4>::
                MArraySlice(*this, r1, r2, r3, r4, NoArgument() );
      }

      template<class T1, class T2, class T3, class T4, class T5>
      typename SliceCounter<T,T1,T2,T3,T4,T5>::MArraySlice
      operator()(T1 r1, T2 r2, T3 r3, T4 r4, T5 r5) const
      {
         return typename SliceCounter<T,T1,T2,T3,T4,T5>::
                MArraySlice(*this, r1, r2, r3, r4, r5 );
      }

      // index with IndexList (rhs version)
      // return MArray<T,1> object
      MArray<T,1> operator()( const IndexList<N>& l ) const
      {
         typename list<IndexV>::const_iterator i=l.begin(), e=l.end();
         MArray<T,1> v( l.size() );

         for( int j=1; i!=e; ++i, ++j )
            v(j) = (*this)( *i );

         return v;
      }

      // (lhs version)
      // asignment to self indexed with IndexList A[l] = Expr<T,1>
      // IndexRef object is needed to carry the operator=( Expr ) method
      IndexRef<T,N> operator[]( const IndexList<N>& l )
      {
         return IndexRef<T,N>( l, *this );
      }

      // Change bases
      void setBase( const int b1 );
      void setBase( const int b1, const int b2 );
      void setBase( const int b1, const int b2, const int b3 );
      void setBase( const int b1, const int b2, const int b3, const int b4 );
      void setBase( const int b1, const int b2, const int b3, const int b4,
                    const int b5 );


      // reference another array
      void makeReference( const MArray<T,N>& other );

      // reallocate memory
      void realloc( const Shape<N>& s );


      // ITERATORS

      // these fulfill requirements of forward iterators at most ...
      iterator iter()
      {
         return iterator( *this );
      }

      iterator begin()
      {
         return iterator( *this );
      }

      const_iterator begin() const
      {
         return const_iterator( *this );
      }

      iterator end()
      {
         return iterator( *this, _iter_end_tag() );
      }

      const_iterator end() const
      {
         return const_iterator( *this, _iter_end_tag() );
      }

      // in case our memory layout is contiguous, we can offer an
      // easy solution to provide random access iterators:
      T* restrict_ beginRA()
      {
         LTL_ASSERT( isStorageContiguous(),
                     "Can't construct random-access iterator on non contiguous MArray!" );
         return data();
      }

      T* restrict_ endRA()
      {
         return beginRA() + nelements();
      }


      // pointer to first data element.
      // NOTE : the second data element NEED NOT be at first + 1
      // since the memeory might be non-contiguous. Use iteators!
      T* restrict_ data() const
      {
         return data_ - shape_.zeroOffset();
      }

      // an iterator-style thing providing indices when dereferenced
      // i.e. *i or i() gives a FixedVector holding the indices and
      // and i(int) gives the index in the i-th dimension
      IndexIterator indexBegin() const
      {
         return IndexIterator( *this );
      }


      // ARRAY INFORMATION, INTERFACES SHAPE

      size_type nelements() const
      {
         return shape_.nelements();
      }

      int minIndex( const int dim ) const
      {
         CHECK_DIM(dim);
         return shape_.base(dim);
      }

      int maxIndex( const int dim ) const
      {
         CHECK_DIM(dim);
         return shape_.last(dim);
      }

      int length( const int dim ) const
      {
         CHECK_DIM(dim);
         return shape_.length(dim);
      }

      int stride( const int dim ) const
      {
         CHECK_DIM(dim);
         return shape_.stride(dim);
      }

      bool isStride1() const
      {
         return stride(1) == 1;
      }

      int zeroOffset() const
      {
         return shape_.zeroOffset();
      }

      bool isStorageContiguous() const
      {
         return shape_.isStorageContiguous();
      }

      bool isConformable( const Shape<N>& other ) const
      {
         return shape_.isConformable( other );
      }

      template<class T2>
      bool isConformable( const MArray<T2,N>& other ) const
      {
         return shape_.isConformable( *other.shape() );
      }

      MemoryBlock<T>* memoryBlock() const
      {
         return memBlock_;
      }

      bool isAllocated() const
      {
         return memBlock_ != NULL;
      }

      const Shape<N>* shape() const
      {
         return &shape_;
      }

      // reorder array (without copy)
      void reverseSelf( const int dim=1 );
      MArray<T, N> reverse( const int dim=1 ) const;
      void transposeSelf( const int dim1=1, const int dim2=2 );
      MArray<T, N> transpose( const int dim1=1, const int dim2=2 ) const;

      // DEBUG OUTPUT
      void describeSelf() const;


   protected:
      // COPY
      template<class T2>
      void copy( const MArray<T2,N>& other );

      void fill( const T t );

      // SUBARRAYS
      void subarray( const MArray<T,N>& other,
                     const Range& r1 );

      void subarray( const MArray<T,N>& other,
                     const Range& r1, const Range& r2 );

      void subarray( const MArray<T,N>& other,
                     const Range& r1, const Range& r2,
                     const Range& r3 );

      void subarray( const MArray<T,N>& other,
                     const Range& r1, const Range& r2,
                     const Range& r3, const Range& r4 );

      void subarray( const MArray<T,N>& other,
                     const Range& r1, const Range& r2,
                     const Range& r3, const Range& r4, const Range& r5 );


      // SLICING (changes rank -> can't be public, use consructors ...)
      template<int N2, class R1, class R2, class R3, class R4, class R5>
      void slicearray( const MArray<T,N2>& other,
                       R1 r1, R2 r2, R3 r3, R4 r4, R5 r5 );

      template<int N2>
      void slice( int& setRank, Range r, const MArray<T,N2>& other,
                  int sourceRank );

      template<int N2>
      void slice( int& setRank, int r, const MArray<T,N2>& other,
                  int sourceRank );

      template<int N2>
      void slice( int& setRank, NoArgument,
                  const MArray<T,N2>& other,
                  int sourceRank )
      { }

      // INTERNAL UTIL METHODS
      void setupMemory();

      void setrange( const int dim, const Range& r );


      // DATA MEMBERS

      MemoryBlock<T>* memBlock_; // our MemoryBlock

      Shape<N>        shape_;    // holds shape information
      // and zeroOffset_ which is the distance between data_ and the
      // the adress of the first element of the array, such that
      // first = data_+sum( base(i) * stride(i) ) = data_ + shape_.zeroOffset()
      // always holds.
   public:
      T* restrict_    data_;     // pointer to element ( 0, 0, ..., 0 )
};


// ====================================================================


#include <ltl/marray/marray_ops.h>
#include <ltl/marray/marray_methods.h>

LTL_END_NAMESPACE

#undef __LTL_IN_FILE_MARRAY__

#endif // __LTL_MARRAY__
