/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: eval.h 23 2003-10-28 02:24:16Z drory $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2002 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */


#ifndef __LTL_IN_FILE_MARRAY__
#error "<ltl/marray/eval.h> must be included via <ltl/marray.h>, never alone!"
#endif


#ifndef __LTL_EVAL_EXPR__
#define __LTL_EVAL_EXPR__


#ifdef LTL_DEBUG_EXPRESSIONS
#include <iostream>
#endif


// this function is called from MArray::operator X= ( TExpr<>& e )
// to actually perform the evaluation and assignment
//
template<class T, int N, class E, class Assigner>
inline void eval_assign_expr( MArray<T,N>& a, TExpr<E,N>& e, Assigner )
{
   CHECK_CONFORM( e, a );

   if( TExpr<E,N>::numIndexIter != 0 )
   {
#ifdef LTL_DEBUG_EXPRESSIONS
      cerr << "found " << TExpr<E,N>::numIndexIter << " index iters\n";
#endif

      return eval_with_index_iter( a, e, Assigner() );
   }

   // if the expression is 1-dimensional or if the memory layout of all
   // operands is contiguous, we can use fast 1-dimensional traversal
   if( N == 1 || (a.isStorageContiguous() && e.isStorageContiguous()) )
      return eval_assign_expr_1( a, e, Assigner() );
   else
      return eval_assign_expr_N( a, e, Assigner() );
}



// we are dealing with the 1 dimensional case
//
template<class T, int N, class E, class Assigner>
void eval_assign_expr_1( MArray<T,N>& a, TExpr<E,N>& e, Assigner )
{
#ifdef LTL_DEBUG_EXPRESSIONS
   cerr << "evaluating with fully collapsed loops\n";
#endif

   const int innerLoopNum = a.nelements();
   const int loopMod4 = innerLoopNum & 3;

   T* restrict_ dataptr = a.data();

   if( a.isStride1() && e.isStride1() )
   {
#ifdef LTL_DEBUG_EXPRESSIONS
      cerr << "evaluating with common stride 1\n";
#endif
      // good, both arrays have stride 1 and both arrays
      // also have contiguous memory
      int j=0;
      for( ; j<loopMod4; ++j )
         Assigner::assign_op( dataptr[j], e.readWithoutStride(j) );

      for( ; j<innerLoopNum; j+=4 )
      {
         // gcc does poor aliasing analysis (in spite of use of restrict),
         // therefore we have to make it clear that writing to the result
         // array does not invalidate _any_ of the data associated with the
         // expression or its iterators (otherwise it reloads the data_
         // pointers of the iterators every time)    :-(
         typename E::value_type tmp1 = e.readWithoutStride(j  );
         typename E::value_type tmp2 = e.readWithoutStride(j+1);
         typename E::value_type tmp3 = e.readWithoutStride(j+2);
         typename E::value_type tmp4 = e.readWithoutStride(j+3);

         Assigner::assign_op( dataptr[j  ], tmp1 );
         Assigner::assign_op( dataptr[j+1], tmp2 );
         Assigner::assign_op( dataptr[j+2], tmp3 );
         Assigner::assign_op( dataptr[j+3], tmp4 );
      }
   }
   else
   {
#ifdef LTL_DEBUG_EXPRESSIONS
      cerr << "evaluating without common stride\n";
#endif
      // well, then slightly less efficient
      const int stride = a.stride(1);
      
      int j=0, k=0;
      for( ; j<loopMod4; ++j, k+=stride )
         Assigner::assign_op( dataptr[j*stride], e.readWithStride(j) );

      for( ; j<innerLoopNum; j+=4, k+=4*stride )
      {
         typename E::value_type tmp1 = e.readWithStride(j  );  
         typename E::value_type tmp2 = e.readWithStride(j+1);
         typename E::value_type tmp3 = e.readWithStride(j+2);
         typename E::value_type tmp4 = e.readWithStride(j+3);

         Assigner::assign_op( dataptr[k         ], tmp1 );
         Assigner::assign_op( dataptr[k+  stride], tmp2 );
         Assigner::assign_op( dataptr[k+2*stride], tmp3 );
         Assigner::assign_op( dataptr[k+3*stride], tmp4 );
      }
   }
}



// now the N-dimensional case
//
template<class T, int N, class E, class Assigner>
void eval_assign_expr_N( MArray<T,N>& a, TExpr<E,N>& e, Assigner )
{
   // we already know that the storage ist not contiguous.
#ifdef LTL_DEBUG_EXPRESSIONS
   cerr << "evaluating with stack traversal\n";
#endif

   const int innerLoopNum = a.length(1);
   const int loopMod4 = innerLoopNum & 3;
   int n = a.nelements();

   typename MArray<T,N>::iterator i = a.begin();

   if( a.isStride1() && e.isStride1() )
   {
#ifdef LTL_DEBUG_EXPRESSIONS
      cerr << "evaluating with common stride 1\n";
#endif

      while( n )
      {
         T* restrict_ dataptr = i.data();

         int j=0;
         for( ; j<loopMod4; ++j )
            Assigner::assign_op( dataptr[j], e.readWithoutStride(j) );

         for( ; j<innerLoopNum; j+=4 )
         {
            // gcc does poor aliasing analysis (in spite of use of restrict),
            // therefore we have to make it clear that writing to the result
            // array does not invalidate _any_ of the data associated with the
            // expression or its iterators (otherwise it reloads the data_
            // pointers of the iterators every time)    :-(
            typename E::value_type tmp1 = e.readWithoutStride(j  );
            typename E::value_type tmp2 = e.readWithoutStride(j+1);
            typename E::value_type tmp3 = e.readWithoutStride(j+2);
            typename E::value_type tmp4 = e.readWithoutStride(j+3);

            Assigner::assign_op( dataptr[j  ], tmp1 );
            Assigner::assign_op( dataptr[j+1], tmp2 );
            Assigner::assign_op( dataptr[j+2], tmp3 );
            Assigner::assign_op( dataptr[j+3], tmp4 );
         }
         i.advanceN( innerLoopNum );
         e.advanceN( innerLoopNum );
         i.advanceDim();
         e.advanceDim();
         n -= innerLoopNum;
      }
   }
   else
   {
#ifdef LTL_DEBUG_EXPRESSIONS
      cerr << "evaluating without common stride\n";
#endif

      const int stride = a.stride(1);

      while( n )
      {
         T* restrict_ dataptr = i.data();

         int j=0, k=0;
         for( ; j<loopMod4; ++j, k+=stride )
            Assigner::assign_op( dataptr[j*stride], e.readWithStride(j) );

         for( ; j<innerLoopNum; j+=4, k+=4*stride )
         {
            typename E::value_type tmp1 = e.readWithStride(j  );  
            typename E::value_type tmp2 = e.readWithStride(j+1);
            typename E::value_type tmp3 = e.readWithStride(j+2);
            typename E::value_type tmp4 = e.readWithStride(j+3);

            Assigner::assign_op( dataptr[k         ], tmp1 );
            Assigner::assign_op( dataptr[k+  stride], tmp2 );
            Assigner::assign_op( dataptr[k+2*stride], tmp3 );
            Assigner::assign_op( dataptr[k+3*stride], tmp4 );
         }
         i.advanceN( innerLoopNum );
         e.advanceN( innerLoopNum );
         i.advanceDim();
         e.advanceDim();
         n -= innerLoopNum;
      }
   }
}


// if we have index expressions involved, we cannot use the above
// optimizations, since we need to keep track of where we are
//
template<class T, int N, class E, class Assigner>
void eval_with_index_iter( MArray<T,N>& a, TExpr<E,N>& e, Assigner )
{
#ifdef LTL_DEBUG_EXPRESSIONS
   cerr << "evaluating with pure stack traversal due to IndexIter\n";
#endif
   // we have an index iterator in the expression,
   // so we cannot do much ...
   typename MArray<T,N>::iterator i = a.begin();
   if( i.isStride1() && e.isStride1() )
   {
      while( !i.done() )
      {
         Assigner::assign_op( *i, *e );
         i.advanceWithStride1();
         e.advanceWithStride1();
         if( i.needAdvanceDim() )
         {
            i.advanceDim();
            e.advanceDim();
         }
      }
   }
   else
   {
      while( !i.done() )
      {
         Assigner::assign_op( *i, *e );
         i.advance();
         e.advance();
         if( i.needAdvanceDim() )
         {
            i.advanceDim();
            e.advanceDim();
         }
      }
   }
}


#endif
