/* 
 * assemble.cpp
 * assemble flagged pixels into objects
 *
 * --------------------------------------------------------------------
 * $Id: assemble.cpp 6 2002-10-29 17:17:10Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include <list>

#include "globals.h"
#include "yoda.h"
#include "object.h"

#include <ltl/marray.h>

//typedef MArray<float,2> MImage;
//typedef MArray<char,2> MMask;
#include <iostream>
#include <iomanip>

using std::cerr;
using std::cout;
using std::setw;

extern Globals g;

void mark_pixels( Object& obj, const  int x, const int y,
		  MMask& M, MImage &I,
		  const int x1, const int x2, const int y1, const int y2 );


list<Object*> assemble( MMask& M, const MImage& I, 
			const MImage& rms, 
			const int border,
			const bool quiet )
{
   if( !quiet )
   {
      cerr << "  Maximum object area: "<< g.maxsize_*g.maxsize_ << " pixels\n";
      cerr << "  Saturation level: " << g.satvalue_ << " ADU\n";
      cerr << "  Minimum number of pixels above threshold: "<<g.npixmin_<<endl;
   }

   list<Object*> lst;

   int objnum = 0, peaknum=0;

   // leave border
   Range R1( I.minIndex(1)+border, I.maxIndex(1)-border );
   Range R2( I.minIndex(2)+border, I.maxIndex(2)-border );
   MImage I2( I, R1, R2 ); 
   MMask  M2( M, R1, R2 ); 
   I2.setBase( R1.first(), R2.first() );
   M2.setBase( R1.first(), R2.first() );

   // have to iterate over all pixels
   MMask::IndexIterator mask = M2.indexBegin();

   while( !mask.done() )
   {
      const int x = mask.index(1);
      const int y = mask.index(2);

      if( M2(x,y) == 1 )
      {
	 // found unprocessed object pixel
	 ++peaknum;

#ifdef YODA_DEBUG_ASSEMBLE
	 cerr << "new obj at (" << x << "," << y << ") : " << peaknum << endl;
#endif

	 // go into recursive fill ..
	 Object *obj = new Object();
	 mark_pixels( *obj, x, y, M2, I2, 
		      R1.first(), R1.last(), R2.first(), R2.last() );
#ifdef YODA_DEBUG_ASSEMBLE
	    cerr << "  marked " << obj->pixnum() << " pixels" << endl;
#endif	  
	 
	 // discard object if too small or too large or saturated ..
	 if( obj->pixnum() >= g.npixmin_ &&
	     obj->pixnum() < g.maxsize_*g.maxsize_ && 
	     obj->maxv() < g.satvalue_ )
	 {
#ifdef YODA_DEBUG_ASSEMBLE
	    cerr << "adding object no. " << objnum << endl;
#endif	  
	    lst.push_front( obj );
	    ++objnum;
	 }
	 else
	 {
#ifdef YODA_DEBUG_ASSEMBLE
	    cerr << "rejecting object no. " << peaknum << endl;
#endif	  
	    delete obj;
	 }
	 if( !(peaknum % 100) && !quiet ) 
	    cerr << setw(7) << peaknum << " "; 
      }
      ++mask;
   }

   if( !quiet )
   {
      cerr << endl;
      cerr << "  "<<peaknum << " peaks above threshold found.\n";
      cerr << "  "<<objnum  << " classified as object, "<< peaknum-objnum
	   << " as noise.\n";
   }

   return lst;
}


// recursively add pixels surrounding a pixel (x,y) to an object
// if they are flagged in the mask as lying above threshold
//
void mark_pixels( Object& obj, const  int x, const int y,
		  MMask& M, MImage &I,
		  const int x1, const int x2, const int y1, const int y2 )
{

   // mark pixel as processed
   M(x,y) = 2;
  
   // add pixel to object
   obj.addpixel( x, y, I(x,y) );

   // recursively mark 8 neighbouring pixels
   if( x+1<=x2 && M(x+1,y) == 1 )
      mark_pixels( obj, x+1, y, M, I, x1, x2, y1, y2 );

   if( x+1<=x2 && y+1<=y2 && M(x+1,y+1) == 1 )
      mark_pixels( obj, x+1, y+1, M, I, x1, x2, y1, y2 );

   if( y+1<=y2 && M(x,y+1) == 1 )
      mark_pixels( obj, x, y+1, M, I, x1, x2, y1, y2 );

   if( x-1>=x1 && y+1<=y2 && M(x-1,y+1) == 1 )
      mark_pixels( obj, x-1, y+1, M, I, x1, x2, y1, y2 );

   if( x-1>=x1 && M(x-1,y) == 1 )
      mark_pixels( obj, x-1, y, M, I, x1, x2, y1, y2 );

   if( x-1>=x1 && y-1>=y1 && M(x-1,y-1) == 1 )
      mark_pixels( obj, x-1, y-1, M, I, x1, x2, y1, y2 );

   if( y-1>=y1 && M(x,y-1) == 1 )
      mark_pixels( obj, x, y-1, M, I, x1, x2, y1, y2 );

   if( x+1<=x2 && y-1>=y1 && M(x+1,y-1) == 1 )
      mark_pixels( obj, x+1, y-1, M, I, x1, x2, y1, y2 );
}







