/*
 * bg.cpp
 * Calculation of background map and rms map of image
 *
 * --------------------------------------------------------------------
 * $Id: bg.cpp 6 2002-10-29 17:17:10Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include <ltl/marray.h>
#include <ltl/marray_io.h>
#include <ltl/statistics.h>

#include "yoda.h"
#include "globals.h"
#include "mathutil.h"

extern Globals g;

using namespace ltl;

void backgroundMap( const MImage& image, MImage& bgrms, MImage& bg, 
		    bool quiet ) 
{
   const int nx = image.length(1);
   const int ny = image.length(2);
   const float nan = g.nanvalue_;

   double irms, iavg;

   if( !quiet )
      cerr << "Background estimation using grid size "<<g.bggrid_<<" ..."<<endl;

   // first guess of mean background and average rms
   kappa_sigma_average( image( Range(1,nx,16), Range(1,ny,16) ), 
                        3.0, nan, &iavg, &irms );

   if( !quiet )
      cerr << "  Average background: " << iavg << ", RMS: " << irms << endl;
  
   MArray<int,1>   h( g.nskybins_ );
   MArray<float,1> b( g.nskybins_ );

   int ngood;

   for( int j=1, yy=1; j<ny; j+=g.bggrid_, ++yy ) 
   {
      const int upy = j+g.bggrid_ > ny ? ny : j+g.bggrid_;
    
      for( int i=1, xx=1; i<nx; i+=g.bggrid_, ++xx ) 
      {
	 const int upx = i+g.bggrid_ > nx ? nx : i+g.bggrid_;
      
	 const MImage w( image, Range(i,upx), Range(j,upy) );
      
	 double avg, rms, avg_save, rms_save;

         ngood = kappa_sigma_average( w, 3.0, nan, &avg, &rms );

	 if( rms != nan ) 
	 {
	    // compute histogram only if it makes sense ...
	    rms_save = rms;
	    avg_save = avg;
	    h = histogram( w, g.nskybins_, 
                           (float)(avg-3*rms), (float)(avg+3*rms), nan );
            
	    getbins( avg-3*rms, avg+3*rms, b );
	    getmaxsigma( h, b, &avg, &rms );
	
	    // if rms changes too much after histogramming, 
	    // something is wrong
	    if( rms/rms_save < 0.5 || rms/rms_save > 2 ) 
	    {
#ifdef YODA_DEBUG_BG
	       cerr << "Histogram looks wierd ... using averages!" << endl;
#endif
	       rms = rms_save;
	       avg = avg_save;
	    }
	 }
	 else
	 {
#ifdef YODA_DEBUG_BG
	       cerr << "All values NaN!" << endl;
#endif
	    rms = 1e10;
	    avg = 1e10;
	 }

#ifdef YODA_DEBUG_BG
	 cerr << xx << "," << yy << " : " << avg << " " << rms << " " << ngood 
	      << " Range: " << i << "," << upx << ":" << j << "," << upy << endl;
#endif
	 //bgrms( Range(i,upx), Range(j,upy) ) = rms;
	 //bg   ( Range(i,upx), Range(j,upy) ) = avg;
	 bgrms( xx, yy ) = (float)rms;
	 bg   ( xx, yy ) = (float)avg;
      }
   }
}

