/*
 * findstars.cpp
 * Create list of bright stars, using sigma(ir1)/<ir1>
 *
 * --------------------------------------------------------------------
 * $Id: findstars.cpp 24 2003-10-29 21:04:29Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 * 
 */

#include "config.h"

#include "globals.h"
#include "yoda.h"
#include "object.h"

#include <ltl/marray.h>
#include <ltl/ascio.h>

#include <list>
#include <functional>

#include <iostream>
#include <iomanip>
using std::setw;

extern Globals g;

struct star_out_of_range : public unary_function<Object *, bool>
{
      star_out_of_range( float lower, float upper ) 
	 : l_(lower), u_(upper)
      {}

      float l_, u_;

      bool operator()( const Object *x ) 
      {
	 if( x->ir1() < l_  || x->ir1() > u_ )
	    return true;
	 return false;
      }
};


// find all bright objects (at least nmin) with rms(ir1)/<ir1> < sigma
//
list<Object*> findstars( list<Object*> l, float sigma )
{
   list<Object*> l1;
  
   // first of all, find highly significant, bright, isolated sources
   for( list<Object*>::iterator i=l.begin(); i!=l.end(); ++i ) 
   {
      Object& obj = **i;
      if( obj.corelum() > g.psfcorelum_*obj.isolum() &&
	  !obj.getsplit() && 
	  obj.asym() <= g.psfmaxasym_ )
	 l1.push_back( &obj );
   }

   // now iterate until no change and nmin ojects have been found
   int n = l1.size();
   float avgir1 = 0.0f;
   float rmsir1 = 0.0f;

   while( n > g.psfminstars_ ) 
   {
      avgir1 = 0.0f;
      rmsir1 = 0.0f;
  
      // compute average and sigma ir1
      for( list<Object*>::iterator i=l1.begin(); i!=l1.end(); ++i )
	 avgir1 += (*i)->ir1();
      avgir1 /= n;

      for( list<Object*>::iterator i=l1.begin(); i!=l1.end(); ++i )
	 rmsir1 += ((*i)->ir1() - avgir1) * ((*i)->ir1() - avgir1);
      rmsir1 = sqrt( rmsir1/(n-1) );
    
      cerr << "  Objects: " << setw(6) << n << ", <ir1> = " << avgir1 
	   << ", rms(ir1) = " << rmsir1 << ", rms/<ir1> = " << rmsir1/avgir1 
	   << endl; 
    
      // stop if rms/avg < sigma
      if( rmsir1/avgir1 < sigma )
	 break;

      // remove objects above avg+3*rms and below avg-1*rms
      l1.remove_if( star_out_of_range(avgir1-3*rmsir1, avgir1+rmsir1) );    

      int n2 = l1.size();

      // break if nothing changes anymore
      if( n2 == n )
	 break;
      n = n2;
   }

   cerr << "  Objects: " << setw(6) << n << ", <ir1> = " << avgir1 
	<< ", rms(ir1) = " << rmsir1 << ", rms/<ir1> = " << rmsir1/avgir1 
	<< endl; 
    
   return l1;
}


list<Object*> readstars( const char *filename, list<Object*> l )
{
   cerr << "Reading catalog of PSF template objects from '"<<filename<<"'.\n";
  
   list<Object*> psflist;
   AscFile f( filename );

   MArray<float,1> x = f.readFloatColumn( 2 );
   MArray<float,1> y = f.readFloatColumn( 3 );
   const int nx = x.length(1);

   for( int j=1; j<=nx; j++ ) 
   {
      for( list<Object*>::iterator i=l.begin(); i!=l.end(); ++i ) 
      {
	 double dx = ((*i)->icx() - x(j));
	 double dy = ((*i)->icy() - y(j));
	 if( dx*dx + dy*dy < 2 )
	    psflist.push_back( *i );
      }
   }
   return psflist;
}


