/* -*- C++ -*-
 * interp.cpp
 * Bilinear interpolation for background and rms images
 *
 * --------------------------------------------------------------------
 * $Id: interp.h 2 2002-08-19 17:00:29Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include <ltl/marray.h>

#include "yoda.h"
#include "globals.h"

extern Globals g;

// f(x,y) = a1 + a2*x + a3*y + a4*x*y
// where the coefficients a1...a4 are determined from f(x0,y0) ... f(x1,y1),
// with (x0,y0), (x1,y0), (x0,y1), and (x1,y1) being the 4 neighbouring
// sampling points

#define CLAMP(x,low,high) (((x)>(high)) ? (high) : (((x)<(low)) ? (low) : (x)))

template<class T, class I1, class I2>
class Interp2DExpr
{
   public: 
      typedef T value_type;
      enum { numIndexIter = 1 };

      Interp2DExpr( const I1& i1, const I2& i2, 
		    const MArray<T,2>& sample, 
		    const int xmesh, const int ymesh )
	 : i1_(i1), i2_(i2), sample_(sample), xmesh_(xmesh), ymesh_(ymesh)
      { }

      Interp2DExpr( const Interp2DExpr<T, I1, I2>& other )
	 : i1_(other.i1_), i2_(other.i2_), 
	   sample_(other.sample_), xmesh_(other.xmesh_), ymesh_(other.ymesh_)
      { }

      T operator*() const
      { return bilin_interp2d(); }

      void operator++()
      { ++i1_; ++i2_; }

      void advance()
      { i1_.advance(); i2_.advance(); }
  
      void advanceN( const int N )
      { } 

      void advanceWithStride1()
      { i1_.advanceWithStride1(); i2_.advanceWithStride1(); }
  
      T readWithStride( const int i ) const
      { return 0; }
         
      T readWithoutStride( const int i ) const
      { return 0; }

      void advanceDim()
      { i1_.advanceDim(); i2_.advanceDim(); }

      bool isStorageContiguous() const
      { return false; }
  
      bool isStride1() const
      { return i1_.isStride1() && i2_.isStride1(); }

      bool isConformable( const Shape<2>& other ) const
      { return i1_.isConformable( other ) && 
	   i2_.isConformable( other ); }

      const Shape<2> *shape() const
      { return i1_.shape(); }

   protected:
      T bilin_interp2d() const;

      I1 i1_;
      I2 i2_;
      const MArray<T,2>& sample_;
      const int xmesh_;
      const int ymesh_;
};


template<class T, class I1, class I2>
T Interp2DExpr<T, I1, I2>::bilin_interp2d() const
{
   const int x=*i1_;
   const int y=*i2_;
 
   // calculate mesh cell in sample_ array
   int xF = (x+xmesh_/2)/xmesh_;
   int yF = (y+ymesh_/2)/ymesh_;

   xF = CLAMP(xF, 1, sample_.maxIndex(1)-1);
   yF = CLAMP(yF, 1, sample_.maxIndex(2)-1);

   const float F00 = sample_(xF   , yF);
   const float F01 = sample_(xF   , yF+1);
   const float F10 = sample_(xF+1 , yF);
   const float F11 = sample_(xF+1 , yF+1);
  
   const float X0 = xF*xmesh_ - xmesh_/2;
   const float X1 = X0 + xmesh_;
   const float Y0 = yF*ymesh_ - ymesh_/2;
   const float Y1 = Y0 + ymesh_;

   const float D = (X1*Y1 - X1*Y0 - X0*Y1 + X0*Y0);
  
   const float a1 = (F00*X1*Y1 - F01*X1*Y0 - F10*X0*Y1 + F11*X0*Y0)/xmesh_/ymesh_;
   const float a2 = (F10*Y1 - F00*Y1 - F11*Y0 + F01*Y0) / D;
   const float a3 = (F01*X1 + F10*X0 - F11*X0 - F00*X1) / D;
   const float a4 = (F11 - F01 + F00 - F10) / xmesh_ / ymesh_;

   // finally return the interpolant
   return a1 + a2*x + a3*y + a4*x*y;
}


template<class T, class I1, class I2>
inline TExpr<Interp2DExpr<T,I1,I2>,2>
interp2d( const I1& i1, const I2& i2, 
	  const MArray<T,2>& a, const int xmesh, const int ymesh  )
{
   typedef Interp2DExpr<T, I1, I2> ExprT;
   return TExpr<ExprT,2>( ExprT( i1,i2, a, xmesh, ymesh ) );
}




