/* -*- C++ -*-
 * mathutil.h
 * Various math utilities
 *
 * --------------------------------------------------------------------
 * $Id: likelihood.h 2 2002-08-19 17:00:29Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */


#ifndef __YODA_LIKELIHOOD__
#define __YODA_LIKELIHOOD__

#include "config.h"

#include <cmath>
#include <ltl/marray.h>

#include "yoda.h"

typedef MArray<float,2> MImage;
typedef MArray<float,1> MVector;


class LiklihoodFunc_Alpha
{
   public:
      LiklihoodFunc_Alpha( const MImage& I, float sky, float lum, 
                           const MImage& psf )
         : I_(I), sky_(sky), lum_(lum), T_(psf)
      { }
      
      float operator()( const float alpha, const float lum  )
      {
         // x(1) = alpha, x(2) = lum
         float l = 0.0, norm=0.0;
         MImage T( T_.shape() );
         for( MImage::IndexIterator i = T.indexBegin(); !i.done(); ++i )
         {
            float tmp = 
               interp2d( T_, float(i(1))/alpha, float(i(2))/alpha );
            tmp *= (tmp>0.f);
            T(i()) = tmp;
            norm += tmp;
         }
         
         for( MImage::IndexIterator i = I_.indexBegin(); !i.done(); ++i )
         {
            const float tmp = I_(i()) / ( fabs(lum)*lum_*T(i())/norm + sky_ );
            l += tmp * (1.-log(tmp)) - 1.;
         }

         const float b1 = fabs(alpha-5.35)/4.65;
         const float b2 = fabs(lum-1.0)/0.75;
         float penalty =  (b1>1 ? exp(pow(b1,4)) : 0.);
         penalty += (b2>1 ? exp(pow(b2,4)) : 0.);

         return -l + penalty;
      }
      
   private:
      const MImage& I_;
      const float sky_, lum_;
      const MImage& T_;
};



class LiklihoodFunc_Beta
{
   public:
      LiklihoodFunc_Beta( const MImage& I, float sky, float lum, 
                          const MImage& psf )
         : I_(I), sky_(sky), lum_(lum), T_(psf)
      { }
      
      float operator()( const float beta, const float lum  )
      {
         // x(1) = alpha, x(2) = lum
         float l = 0.0, norm=0.0;
         MImage T( T_.shape() );
         for( MImage::IndexIterator i = T.indexBegin(); !i.done(); ++i )
         {
            float tmp = (1-beta) * T_(i()) + 
               beta*interp2d( T_, float(i(1))/2.f, float(i(2))/2.f );
            tmp *= (tmp>0.f);
            T(i()) = tmp;
            norm += tmp;
         }
         
         for( MImage::IndexIterator i = I_.indexBegin(); !i.done(); ++i )
         {
            const float tmp = I_(i()) / ( fabs(lum)*lum_*T(i())/norm + sky_ );
            l += tmp * (1.-log(tmp)) - 1.;
         }
         const float b1 = fabs(beta-0.5)/0.5;
         const float b2 = fabs(lum-1.0)/0.5;
         float penalty =  (b1>1 ? exp(pow(b1,4)) : 0.);
         penalty += (b2>1 ? exp(pow(b2,4)) : 0.);

         return -l + penalty;
      }
      

   private:
      const MImage& I_;
      const float sky_, lum_;
      const MImage& T_;
};


#endif
