/* 
 * main.cpp
 * main entry and command line processing
 *
 * --------------------------------------------------------------------
 * $Id: main.cpp 42 2014-09-24 22:58:44Z drory $
 * --------------------------------------------------------------------
 *
 * Copyright (C) 2000-2001 Niv Drory <drory@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 */

#include "config.h"

#include <list>
#include <exception>

#include <ltl/util/option.h>
#include <ltl/util/option_parser.h>
#include <ltl/util/command_line_reader.h>

using namespace util;
using namespace std;

#include "globals.h"
#include "yoda.h"

Globals g;
char* version_str = "1.2.7 2014-09-24";


void setupOptions( OptionParser& op );


int main( int argc, char *argv[] )
{
   // setup program options ...
   CommandLineReader& clr = *new CommandLineReader( argc, argv );
   OptionParser& op = *new OptionParser( &clr );
   setupOptions( op );

   // parse command line ...
   try 
   {
      op.parseOptions();
   }
   catch( UException& e ) 
   {
      cerr << "Caught exception : " << e.what() << endl << endl;
      cout << "YODA " << version_str << endl << endl;
      cerr << "Usage: yoda [options ...] FITS-file \n\n";
      op.printUsage( cerr );
      cerr << endl;
      exit(1);
   }

   // check for --version or -V
   if( g.version_ )
   {
      cout << "YODA " << version_str << endl << endl;
      cout << "Copyright (C) Niv Drory" << endl;
      cout << "YODA comes with ABSOLUTELY NO WARRANTY." << endl;
      cout << "You may redistribute copies of YODA under the terms of"<<endl;
      cout << "the GNU General Public License. See the file named COPYING." 
	<< endl << endl;
      cout << "Written by Niv Drory"<< endl;
      exit(0);
   }

   // check for --help
   if( g.help_ )
   {
      cout << "YODA " << version_str << endl << endl;
      cout << "YODA (Yet another Object Detection Application) is a software package\nperforming object detection, photometry and star-galaxy classification\non astronomical images. " << endl << endl;
      
      cout << "Usage: yoda [options ...] FITS-file \n\n";
      op.printUsage( cout );
      cout << endl << endl;
      cout << "Examples: " << endl;
      cout << "Detect objects in 'image.fits', write catalog to 'master.cat':"<<endl;
      cout << "yoda -D -d master.cat -t 3.0 -n 7 -s 3.0 -z 23.43 image.fits" << endl<<endl;
      cout << "Photometry using detected objects, output to image.phot:"<<endl;
      cout << "yoda -P -p image.phot -m master.cat -z 23.43 image.fits" <<endl<< endl;
      cout << "Create a PSF template from stars in 'stars.list':"<<endl;
      cout << "yoda -T -i stars.list --psf-size 11 image.fits" << endl<<endl;
      cout << "Classify objects:"<<endl;
      cout << "yoda -C -c image.class --psf-input-file image.psf -m image.phot"<<endl<< endl;     
      cout << "It is also possible to run multiple phases (detection, photometry, classification) in one yoda call." << endl;
      cout << endl;
      cout << "Report bugs to <drory@astro.as.utexas.edu>" << endl << endl;
      exit(0);
   }

   // check that we have at least one file to process ...
   list<string> files = clr.fileArguments();
   if( files.empty() && !g.version_ ) 
   {
      cout << "YODA " << version_str << endl << endl;
      cerr << "Usage: yoda [options ...] FITS-file\n";
      cerr << "       run yoda --help for a list of options\n\n";
      exit(1);
   }
  
  
   // ok, now go for it ...
   try 
   {
      g.argc_ = argc;
      g.argv_ = argv;
      
      // We offer the following modes of operation
      //  o Detection
      //  o PSF template generation
      //  o Photometry
      //  o Classification

      // the basic setup is the following:
      //
      //  o only detection is OK.
      // 
      //  o only photometry is also OK, and if we are driven
      //    by a master catalog, no detection pass has to be run
      // 
      //  o Detection and photometry (and PSF and/or classification) is 
      //    also fine
      //
      //  o if we should classify or construct a PSF template, we MUST
      //    also detect (to get object properties for the PSF, or
      //    accurate centroids for comparison with templates for
      //    classification, even if we classify objects driven by a master
      //    catalog

      if( g.detect_ || g.psf_ || g.class_ ) 
      {
	 g.infile_ = files.front();
	 files.pop_front();
	 detect();
      }

      // generate psf template
      if( g.psf_ ) 
      {
	 makepsf( g.Image_, g.objlist_ );
	 write_psffile( g.psfimage_ );
      }

      // write object catalogue
      if( g.detect_ )
	 write_master_catalog( g.objlist_, g.objcat_ );

      // perform photometry
      if( g.photometry_ )
      {
	 if( !g.detect_ ) 
	 {
	    g.infile_ = files.front();
	    files.pop_front();
            const bool global_sky = (g.skyann_[0]<0.5) && (g.skyann_[1]<0.5);
            if( global_sky )
               determine_sky();
	 }
         
	 photometry();
         write_photometry_catalog( g.photlist_, g.photcat_ );
      }
      
      // perform classification
      if( g.class_ )
      {
	 classify();
         write_classify_catalog( g.photlist_, g.classcat_ );
      }
   }
   catch( exception& e ) 
   {
      cerr << "Caught exception : " << e.what() << endl << endl;
      exit(1);
   }
}


void setupOptions( OptionParser& op )
{
   try 
   {
      op.addOption( new BoolOption( "help", "false", 
         "Print this help message and exit", 'h' , &g.help_) );
      op.addOption( new BoolOption( "version", "false", 
         "Print version information and exit", 'V' , &g.version_) );
      // detection
      op.addOption( new BoolOption( "detect", "false", 
         "Execute detection pass", 'D' , &g.detect_) );
      op.addOption( new StringOption( "detect-cat", "<Image>.cat", 
         "Name of output master catalog", 'd', &g.objcat_) );
      op.addOption( new FloatOption( "seeing", "2.0", 
         "FWHM of Gaussian convolution kernel in pixels", 's' , &g.seeing_) );
      op.addOption( new FloatOption( "threshold", "3.0", 
	 "Threshold for object detection above local noise", 't' , &g.thresh_) );
      op.addOption( new IntOption( "npixmin", "7", 
	 "Minimum  number of consecutive pixels above threshold for object detection", 'n', &g.npixmin_) );
      op.addOption( new IntOption( "maxsize", "100", 
	 "Maximum linear size of object", 0, &g.maxsize_) );
      op.addOption( new FloatOption( "nan-value", "0.0", 
	 "Value indicating UNDEFINED pixels in image", 'N' , &g.nanvalue_) );
      op.addOption( new FloatOption( "saturation", "1e6", 
	 "Saturation value of image", 'S' , &g.satvalue_) );
      op.addOption( new FloatOption( "zeropoint", "23.5", 
	 "Magnitude zero-point",'z' , &g.magzero_) );
      op.addOption( new IntOption( "bg-gridsize", "128", 
	 "Number of bins in background histograms", 'g', &g.bggrid_) );
      op.addOption( new IntOption( "bg-bins", "25", 
	 "Number of bins in background histograms", 0, &g.nskybins_) );
      op.addOption( new BoolOption( "bg-iter", "false",
	 "Iterate background and rms estimation", 0, &g.bg_iter_ ) );
      op.addOption( new BoolOption( "dump-bg", "false",
	"Dump background and rms images to disk", 0, &g.bg_dump_images_ ) );
      op.addOption( new BoolOption( "dump-detection", "false",
	 "Dump detection mask to disk", 0, &g.bg_dump_mask_ ) );
      op.addOption( new IntOption( "split-npixmin", "4", 
	 "Minimum  number of pixels in splitted objects",0 , &g.splitnpix_) );
      op.addOption( new FloatArrayOption( "split-range", "0.01,0.8", 
         "Minimum and maximum isophote in units of object maximum flux to split at",
                                          0 , 2, &(g.splitrange_[0])) );
      op.addOption( new FloatOption( "step-split", "0.02", 
	 "Fractional increment in splitting isophote",0 , &g.fracsplit_) );
    
      // PSF
      op.addOption( new BoolOption( "psf", "false", 
	 "Generate PSF template image from objects in image", 'T', &g.psf_ ) );
      op.addOption( new StringOption( "psf-output-file", "<Image>.psf", 
	 "Name of output PSF image file",0 , &g.psffile_) );
      op.addOption( new StringOption( "psf-input-cat", "", 
	 "Read catalog of objects for PSF estimation from file", 'i',
				      &g.psfinputcat_) );
      op.addOption( new StringOption( "psf-catalog", "psf.cat", 
	 "Write catalog of automatically determined PSF objects to file",'I',
                                      &g.psfstarcat_) );
      op.addOption( new FloatOption( "psf-corelum", "10.0", 
	 "Min core luminosity above detection isophote for PSF objects",0,
				     &g.psfcorelum_) );
      op.addOption( new FloatOption( "psf-maxasym", "0.05", 
	 "Maximum asymmetry for PSF objects",0 , &g.psfmaxasym_) );
      op.addOption( new FloatOption( "psf-rms-ir1", "0.1", 
	 "rms(ir1)/<ir1> for candidate stars in PSF estimate",0, &g.psfrmsir1_) );

      op.addOption( new IntOption( "psf-minstars", "11", 
	 "Minimum number of objects to use in PSF estimation",0, 
				   &g.psfminstars_) );
      op.addOption( new IntOption( "psf-size", "9", 
	 "Size of desired PSF template image", 'q', &g.psftemplsize_) );

      // photometry
      op.addOption( new BoolOption( "photometry", "false", 
	 "Perform aperture photometry", 'P', &g.photometry_ ) );
      op.addOption( new StringOption( "photometry-cat", "<Image>.p", 
	 "Name of output photometry catalog",'p' , 
				      &g.photcat_ ) );
      op.addOption( new StringOption( "master-cat", "", 
	 "File of yoda catalog of master image when performing photometry (none: use objects from current detection pass; needed for aperture geometry), and file containing photometry data when performing classification (needed for sky value)",
                                      'm', &g.mastercat_ ) );
      op.addOption( new StringOption( "coords-cat", "", 
	 "File with aperture centers for photometry (none: use coordinates from  master-cat)", 'M' , &g.coordscat_ ) );
      op.addOption( new FloatArrayOption( "coords-columns", "2,3", 
	 "Column numbers of x and y coordinates in coords-cat file",0, 2, 
                                      &(g.coordscolumns_[0]) ) );
      op.addOption( new BoolOption( "no-kron-ap", "false", 
	 "Don't perform kron-aperture photometry", 0, &g.no_kron_ap_ ) );
      op.addOption( new FloatOption( "fixed-aper", "0.0",
	 "Fixed circular aperture diameter in pixels", 'A', 
					  &g.fixed_aper_ ) );
      op.addOption( new FloatOption( "kron-factor", "2.5", 
	 "Size factor for Kron-like apertures", 'k', &g.kronfactor_) );
      op.addOption( new FloatOption( "min-aperture", "2.0", 
	 "Radius of smallest possible kron aperture in pixels", 0, &g.minaperture_) );
      op.addOption( new FloatArrayOption( "ap-transform", "1,0,0,1",
	 "Linear transforamtion coefficients for apertures", 0, 4, 
					  &(g.aptransf_[0]) ));      
      op.addOption( new BoolOption( "iraf-transform", "false", 
	 "Aperture lin. Tranformation in iraf format: xmag,ymag,xrot,yrot", 0, &g.iraftrans_ ) );
      op.addOption( new FloatArrayOption( "sky-annulus", "5,3",
	 "Distance from largest aperture and width of sky annulus. If 0,0 then the sky at each object's position is estimated by interpolating the background image at the object's center.", 0, 2, 
					  &(g.skyann_[0]) ));
      op.addOption( new StringOption( "gain-keyword", "GAIN    " , 
	 "FITS header keyword for GAIN",0 , &g.gainkey_) );
      op.addOption( new FloatOption( "gain", "0.0", 
	 "CCD gain in e-/ADU (0.0 for GAIN FITS keyword)",0 , &g.gain_) );

      // classification
      op.addOption( new BoolOption( "classify", "false", 
                                    "Perform classification", 'C',
                                    &g.class_ ) );
      op.addOption( new StringOption( "psf-input-file", "", 
	 "File containing PSF template (none for internally generated template)",0 , &g.psftemplate_ ) );
      op.addOption( new StringOption( "class-output-cat", "<Image>.class", 
	 "Name of classifier outpout file",'c' , &g.classcat_ ) );
   }
   catch( UException e ) 
   {
      cerr << "Caught exception : " << e.what() << endl;
      exit(1);
   }
}

